/*-
 * Copyright (c) 2005 - 2011 CAS Dev Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CAS Dev. Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      ASMultipartParser.pp
 *
 * $CAS$
 */

// C++ Includes
#include "ASMultipartParser.hpp"

#include "ASLogger.hpp"
#include "ASSAPIUtil.hpp"

#include <errno.h>
#include <stdio.h>
#include <string.h>

namespace CAS // C++ Application Server
{

#define C_ESCAPE_BUFFER_LEN 8192


#define C_CONTENT_DISPOSITION           "Content-Disposition"
#define C_CONTENT_TYPE                  "Content-Type"
#define C_CONTENT_TRANSFER_ENCODING     "Content-Transfer-Encoding"

#define C_KEY_CONTENT_DISPOSITION       0x00000001
#define C_KEY_CONTENT_TYPE              0x00000002
#define C_KEY_CONTENT_TRANSFER_ENCODING 0x00000003

// Undefined state of parser
#define C_UNDEF                0x00000000
// End of MIME message
#define C_END_MIME             0x01000000
// Unrecoverable error
#define C_ERROR                0x10000000


// Boundary section

#define C_BOUNDARY_IN          0x00010001

#define C_BOUNDARY_IN_SUFFIX1  0x00010020
#define C_BOUNDARY_IN_SUFFIX2  0x00010040
#define C_BOUNDARY_IN_SUFFIX3  0x00010080
#define C_BOUNDARY_IN_SUFFIX4  0x00010100

// Header section
#define C_HEADER               0x00020000
#define C_HEADER_CR            0x00020001
#define C_HEADER_LF            0x00020002

#define C_HEADER_CR2           0x00020004
#define C_HEADER_LF2           0x00020008

// Header fields
#define C_KEY_NAME             0x00020010
#define C_KEY_PRE_VALUE        0x00020020
#define C_KEY_TAB_VALUE        0x00020040
#define C_KEY_VALUE            0x00020100

// Body section
#define C_BODY                 0x00040000


#define C_BODY_FILE            0x00000001
#define C_BODY_VALUE           0x00000002

#define C_FIELD_IS_NAME        0x00100000
#define C_FIELD_IS_FILENAME    0x00200000

#define C_TMP_FNAME_SIZE       1024

//
// A constructor
//
MultipartParser::MultipartParser(CTPP::CDT                   & oIData,
                                 ASFilePool                  & mIFiles,
                                 STLW::vector<STLW::string>  & vIFilelist,
                                 const STLW::string          & sIUploadDir,
                                 CCHAR_P                       szIBoundary,
                                 const INT_64                & iIMaxFileSize,
                                 ASServer                    & oIServer,
                                 ASServer::ASRequestContext  & oIRequestContext,
                                 ASRequest                   & oIRequest,
                                 ASLogger                    & oILogger): oData(oIData),
                                                                          mFiles(mIFiles),
                                                                          vFilelist(vIFilelist),
                                                                          sUploadDir(sIUploadDir),
                                                                          szBoundary(szIBoundary),
                                                                          szBoundaryPos(szBoundary + 2),
                                                                          iMaxFileSize(iIMaxFileSize),
                                                                          iState(C_BOUNDARY_IN),
                                                                          iValueState(C_UNDEF),
                                                                          oServer(oIServer),
                                                                          oRequestContext(oIRequestContext),
                                                                          oRequest(oIRequest),
                                                                          oLogger(oILogger)
{
	// Add slash, if need
	if (sUploadDir.size() > 0 && sUploadDir[sUploadDir.size() - 1] != '/') { sUploadDir.append("/", 1); }
}

//
// Parse HTTP header
//
void MultipartParser::ParseHeader()
{
	CCHAR_P szBuffer = sBuffer.c_str();
	// form-data; name="submit-name"
	// form-data; name="files"; filename="c:\text.txt"
	// form-data
	// 123456789
	CCHAR_P szData    = StrFirstCaseStr(szBuffer, "form-data");
	CCHAR_P szDataEnd = szBuffer + sBuffer.size();

	// Check parameter name
	if (szData != (szBuffer + 9)) { return; }

	// Check ';'
	if (szData == szDataEnd || *szData != ';') { return; }
	++szData;

	// Parse string
	for (;;)
	{
		// Skip spaces
		while (szData != szDataEnd && *szData == ' ') { ++szData; }

		INT_32 iKeyName = C_UNDEF;
		CCHAR_P szKeyStart = szData;
		// Find '='
		while (szData != szDataEnd && *szData != '=') { ++szData; }

		// Name
		if (StrNFirstCaseStr(szKeyStart, szData - szKeyStart, "name") != NULL)
		{
			iKeyName = C_FIELD_IS_NAME;
			iValueState = C_BODY_VALUE;
		}
		// FileName
		else if (StrNFirstCaseStr(szKeyStart, szData - szKeyStart, "filename") != NULL)
		{
			iKeyName = C_FIELD_IS_FILENAME;
			iValueState = C_BODY_FILE;
		}

		if (szData == szDataEnd || *szData != '=') { break; }
		++szData;
		if (szData == szDataEnd || *szData != '"') { break; }
		++szData;
		szKeyStart = szData;
		// Find '"'
		while (szData != szDataEnd && *szData != '"') { ++szData; }
		if (szData == szDataEnd) { break; }

		// Store field name
		if      (iKeyName == C_FIELD_IS_NAME) { oFile.name.assign(szKeyStart, szData - szKeyStart); }
		// Store & process filename
		else if (iKeyName == C_FIELD_IS_FILENAME)
		{
			// Store filename with full path
			oFile.full_filename.assign(szKeyStart, szData - szKeyStart);

			// Nothing to do
			if (oFile.full_filename.empty()) { F = NULL; }
			// Write file
			else
			{
				CCHAR_P szFileName = szData;
				// Store filename without path
				for (;;)
				{
					if (szFileName == szKeyStart) { break; }
					if (*szFileName == '/' || *szFileName == '\\') { ++szFileName; break; }
					--szFileName;
				}
				oFile.filename.assign(szFileName, szData - szFileName);
				oFile.filesize = 0;

				// Create temp file name
				CHAR_8 szTMPFileName[C_TMP_FNAME_SIZE];
				snprintf(szTMPFileName, C_TMP_FNAME_SIZE, "u%d.p%d.t%d.r%d", int(getuid()), int(getpid()), int(time(NULL)), int(rand()));
				if (*szTMPFileName == '\0') { F = NULL; }
				else
				{
					oFile.tmp_name.assign(sUploadDir);
					oFile.tmp_name.append(szTMPFileName);

					// Open file for writing
					F = fopen(oFile.tmp_name.c_str(), "wb");
					if (F != NULL) { vFilelist.push_back(oFile.tmp_name); }
					else
					{
						STLW::string sTMP("CAS: Cannot open file \"");
						sTMP.append(oFile.tmp_name);
						sTMP.append("\" for writing: ");
						sTMP.append(strerror(errno));
						oLogger.WriteLog(AS_LOG_WARNING, sTMP.c_str());
					}
				}
			}
		}
		++szData;
		if (szData == szDataEnd || *szData != ';') { break; }
		++szData;
		if (szData == szDataEnd) { break; }
	}

	oServer.HandleRFC1867File(oRequestContext, oFile.name, oFile.full_filename, oFile.filename, oFile.tmp_name, oRequest, &oLogger);
/*
fprintf(stderr, "NAME          `%s`\n", oFile.name.c_str());
fprintf(stderr, "FULL FILENAME `%s`\n", oFile.full_filename.c_str());
fprintf(stderr, "FILENAME      `%s`\n", oFile.filename.c_str());
fprintf(stderr, "TMP NAME      `%s`\n\n\n", oFile.tmp_name.c_str());
fflush(stderr);
*/
}

//
// Handle end of buffer
// If szStringSave pointer is NOT NULL, and
// 1. state == C_BODY_FILE, write data to the disc
// 2. state == C_BODY_VALUE, add data to buffer
//
void MultipartParser::HandleEndOfBuffer(CCHAR_P           & szStringSave,
                                        CCHAR_P             szBoundaryStart,
                                        const UINT_32     & iNewState)
{
	// Store szBoundary, if need
	if (szStringSave != NULL)
	{
		// Write data to file
		if (iValueState == C_BODY_FILE)
		{
			// Avoid stupid warning "warn_unused_result"
			size_t iHandleBytes = 0;
			// Write boundary-like data
			if (szPrevBlockBoundary != NULL)
			{
				const UINT_32 iBufSize = szPrevBlockBoundary - szBoundary;
				if (F != NULL)
				{
					oFile.filesize += iBufSize;
					// Don't write data if file size is greater than iMaxFileSize
					if (!(iMaxFileSize != -1 && oFile.filesize >= UINT_32(iMaxFileSize))) { iHandleBytes = fwrite(szBoundary, 1, iBufSize, F); }
				}
				szPrevBlockBoundary = NULL;
			}

			// Write data
			UINT_32 iBufSize = szBoundaryStart - szStringSave;
			if (F != NULL && iBufSize != 0)
			{
				oFile.filesize += iBufSize;
				// Don't write data if file size is greater than iMaxFileSize
				if (!(iMaxFileSize != -1 && oFile.filesize >= UINT_32(iMaxFileSize))) { iHandleBytes = fwrite(szStringSave, 1, iBufSize, F); }
			}
		}
		// Append buffer
		else if (iValueState == C_BODY_VALUE)
		{
			// Write boundary-like data
			if (szPrevBlockBoundary != NULL)
			{
				const UINT_32 iBufSize = szPrevBlockBoundary - szBoundary;
				sBuffer.append(szBoundary, iBufSize);
				szPrevBlockBoundary = NULL;
			}

			UINT_32 iBufSize = szBoundaryStart - szStringSave;
			sBuffer.append(szStringSave, iBufSize);
		}
	}
	// Set new state
	iState       = iNewState;
	// Store buffer position
	szStringSave = szBoundaryStart;
}

//
// Commit operation
//
void MultipartParser::CommitSection()
{
	// Write data to file
	if (iValueState == C_BODY_FILE)
	{
		// Only if file was open
		if (F != NULL)
		{
			fclose(F);

			// Unlink file if file size is greater than iMaxFileSize
			if (iMaxFileSize != -1 && oFile.filesize >= UINT_64(iMaxFileSize))
			{
				CHAR_8 szBuffer[128];

				STLW::string sTMP("File \"");
				sTMP.append(oFile.filename);
				sTMP.append("\" not saved: size ");
				snprintf(szBuffer, 127, "%llu", (unsigned long long) (oFile.filesize));
				sTMP.append(szBuffer);
				sTMP.append(" is larger than max allowed (");
				snprintf(szBuffer, 127, "%llu", (unsigned long long) (iMaxFileSize));
				sTMP.append(szBuffer);
				sTMP.append(") bytes");

				oLogger.WriteLog(AS_LOG_WARNING, sTMP.c_str());

				unlink(oFile.tmp_name.c_str());
				oFile.tmp_name.erase();
			}
			// Store name in pool
			mFiles.AddFile(oFile);
		}

		// Clear data
		oFile.name.erase();
		oFile.full_filename.erase();
		oFile.filename.erase();
		oFile.tmp_name.erase();
	}
	// Append buffer
	else if (iValueState == C_BODY_VALUE)
	{
		StorePair(oData, oFile.name, sBuffer);
	}
}

//
// Parse chunk of data
//
void MultipartParser::ParseChunk(CCHAR_P szString, CCHAR_P szStringEnd)
{
	CCHAR_P szBoundaryStart = szString;
	CCHAR_P szStringSave    = szString;

	if (szBoundaryPos != szBoundary) { szPrevBlockBoundary = szBoundaryPos; }
	else                             { szPrevBlockBoundary = NULL;          }

	//
	// \r - \n - dash - dash - BOUNADRY [dash - dash] - \r - \n
	// +------------------------------+  |      |       |    |
	// |                                 |      |       |    |
	// |                                 |      |       |    BOUNDARY_IN_SUFFIX4
	// |                                 |      |       |
	// |                                 |      |       BOUNDARY_IN_SUFFIX3
	// |                                 |      |
	// |                                 |      BOUNDARY_IN_SUFFIX2
	// |                                 |
	// |                                 BOUNDARY_IN_SUFFIX1
	// |
	// C_BOUNDARY_IN


	if (iState == C_BOUNDARY_IN)         { goto BOUNDARY_IN;         }

	if (iState == C_BOUNDARY_IN_SUFFIX1) { goto BOUNDARY_IN_SUFFIX1; }
	if (iState == C_BOUNDARY_IN_SUFFIX2) { goto BOUNDARY_IN_SUFFIX2; }
	// Never happenes, but let it still here
	if (iState == C_BOUNDARY_IN_SUFFIX3) { goto BOUNDARY_IN_SUFFIX3; }
	if (iState == C_BOUNDARY_IN_SUFFIX4) { goto BOUNDARY_IN_SUFFIX4; }

	// Header: [space] HeaderValue; optional="field" \r - \n [ \r - \n ]
	// |        |      |                              |    |    |    |
	// |        |      |                              |    |    |    C_HEADER_LF2
	// |        |      |                              |    |    |
	// |        |      |                              |    |    HEADER_CR2
	// |        |      |                              |    |
	// |        |      |                              |    HEADER_LF
	// |        |      |                              |
	// |        |      |                              HEADER_CR
	// |        |      |
	// |        |      C_KEY_VALUE
	// |        |
	// |        C_KEY_PRE_VALUE
	// |
	// C_KEY_NAME

	if (iState == C_HEADER)     { goto HEADER;     }
	if (iState == C_HEADER_CR)  { goto HEADER_CR;  }
	if (iState == C_HEADER_LF)  { goto HEADER_LF;  }
	if (iState == C_HEADER_CR2) { goto HEADER_CR2; }
	if (iState == C_HEADER_LF2) { goto HEADER_LF2; }

	if (iState == C_KEY_NAME)      { goto KEY_NAME;      }
	if (iState == C_KEY_PRE_VALUE) { goto KEY_PRE_VALUE; }
	if (iState == C_KEY_VALUE)     { goto KEY_VALUE;     }

	for(;;)
	{

		// Check szBoundary
		szBoundaryPos = szBoundary;
		// Store state
		iPrevState   = iState;

		// Start of szBoundary
		iState = C_BOUNDARY_IN;
BOUNDARY_IN:
		// Start of szBoundary
		szBoundaryStart = szString;

//fprintf(stderr, "BOUNDARY_IN\n");
		// Boundary body
		for(;;)
		{
//fprintf(stderr, "0x%02X <> 0x%02X\n", *szString, *szBoundaryPos);
			// End of szBoundary found
			if (*szBoundaryPos == '\0')
			{
				szPrevBlockBoundary = NULL;
//fprintf(stderr, "BUFFER: `%s`\n", szString);
				// End of buffer
				if (szString == szStringEnd) { HandleEndOfBuffer(szStringSave, szBoundaryStart, C_BOUNDARY_IN); return; }

				// Boundary suffix
				if(*szString == '-')
				{
//fprintf(stderr, "BOUNDARY_IN_SUFFIX1\n");
					// End of buffer?
					++szString;
BOUNDARY_IN_SUFFIX1:
					if (szString == szStringEnd) { HandleEndOfBuffer(szStringSave, szBoundaryStart, C_BOUNDARY_IN_SUFFIX1); return; }

					// Fatal error: double dashes not found after szBoundary
					if(*szString != '-') { iState = C_ERROR; return; }
//fprintf(stderr, "BOUNDARY_IN_SUFFIX2\n");
					// End of buffer?
					++szString;
BOUNDARY_IN_SUFFIX2:
					if (szString == szStringEnd) { HandleEndOfBuffer(szStringSave, szBoundaryStart, C_BOUNDARY_IN_SUFFIX2); return; }

					// End of MIME reached, no errors
					iState = C_END_MIME;
				}
BOUNDARY_IN_SUFFIX3:
//fprintf(stderr, "BOUNDARY_IN_SUFFIX3\n");
				// CRLF after szBoundary, REQUIRED in all cases!
				if(*szString != '\r') { iState = C_ERROR; return; }

				++szString;
				// End of buffer
				if (szString == szStringEnd) { HandleEndOfBuffer(szStringSave, szBoundaryStart, iState | C_BOUNDARY_IN_SUFFIX4); return; }
BOUNDARY_IN_SUFFIX4:
//fprintf(stderr, "BOUNDARY_IN_SUFFIX4\n");
				// CRLF after szBoundary, REQUIRED in all cases!
				if(*szString != '\n') { iState = C_ERROR; return; }

				++szString;
				// End of buffer
				if (szString == szStringEnd)
				{
					HandleEndOfBuffer(szStringSave, szBoundaryStart, iState | C_HEADER);
					CommitSection();

					// End of MIME
					if ((iState & C_END_MIME) == C_END_MIME)
					{
						// End of MIME, end of processing
						iState = C_END_MIME;
//fprintf(stderr, "->>>>>>>>>>>>>>>>>>> End of MIME, end of processing\n");
					}
					else
					{
//fprintf(stderr, "->>>>>>>>>>>>>>>>>>> End of MIME, start of header\n");
						// New state if HEADER
						iState = C_HEADER;
					}
					return;
				}
//fprintf(stderr, "->>>>>>>>>>>>>>>>>>> End of part of MIME, start of header\n");

//fprintf(stderr, "szString = `%s`\n", szString);
				HandleEndOfBuffer(szStringSave, szBoundaryStart, C_HEADER);
				// Commit operation
				CommitSection();

				iState = C_HEADER;
				break;
			}

			// Boundary not found
			if (*szBoundaryPos != *szString)
			{
//fprintf(stderr, "MIME mistake, revert to old state\n");
//fprintf(stderr, "%c <> %c\n", *szString, *szBoundaryPos);
				// Write data and rollback to previous state
				// Not need here?
				// HandleEndOfBuffer(szStringSave, szBoundaryStart, C_UNDEF);
				// Write tail of MIME, if state is

				iState        = iPrevState;
				szBoundaryPos = szBoundary;
				break;
			}

			// Next position
			++szString;
			++szBoundaryPos;

			// End of buffer, write data
			if (szString == szStringEnd)
			{
				// Store pre-szBoundary data
				HandleEndOfBuffer(szStringSave, szBoundaryStart, C_BOUNDARY_IN);
				return;
			}
		}

//fprintf(stderr, "*!!!!!!!!!!!!!!!!!!! 0x%08X  <> 0x%08X <> 0x%08X\n", iState, (iState & C_HEADER), C_HEADER);
		// Boundary end, start handling of headers section
		if ((iState & C_HEADER) == C_HEADER)
		{
//fprintf(stderr, "C_HEADER *szString = `%s`\n", szString);
			// Handle headers section
			for(;;)
			{
HEADER:
				// Clear temp. buffer
				sBuffer.erase();
				iState = C_KEY_NAME;
				szStringSave = szString;

HEADER_CR:
KEY_NAME:
KEY_PRE_VALUE:
KEY_VALUE:
				// Handle header line
				for (;;)
				{
//fprintf(stderr, "*szString = `%s`\n", szString);
					// Header value
					if (*szString == ':' && ((iState & C_KEY_NAME) == C_KEY_NAME))
					{
						sBuffer.append(szStringSave, szString - szStringSave);

						// Process key name
						CCHAR_P szHeaderName = sBuffer.c_str();
//fprintf(stderr, "szHeaderName = `%s`\n", szHeaderName);

						// Check "Content-Disposition" header
						if (StrFirstCaseStr(szHeaderName, C_CONTENT_DISPOSITION) != NULL)
						{
							iHeaderKeyId = C_KEY_CONTENT_DISPOSITION;
//fprintf(stderr, "Found: " C_CONTENT_DISPOSITION "\n");
						}
						// Check "Content-Type" header
						else if (StrFirstCaseStr(szHeaderName, C_CONTENT_TYPE) != NULL)
						{
//fprintf(stderr, "Found: " C_CONTENT_TYPE "\n");
							iHeaderKeyId = C_KEY_CONTENT_TYPE;
						}
						// Check "Content-Transfer-Encoding"
						else if (StrFirstCaseStr(szHeaderName, C_CONTENT_TRANSFER_ENCODING) != NULL)
						{
//fprintf(stderr, "Found: " C_CONTENT_TRANSFER_ENCODING "\n");
							iHeaderKeyId = C_KEY_CONTENT_TRANSFER_ENCODING;
						}
						else
						{
//fprintf(stderr, "Found: Unknown key\n");
							iHeaderKeyId = 0;
						}

						// Free buffer
						sBuffer.erase();

						// Store start point of string
						szStringSave         = szString;
						iState = C_KEY_PRE_VALUE;
					}
					// Header key => value delimiter
					else if (*szString != ' ' && ((iState & C_KEY_PRE_VALUE) == C_KEY_PRE_VALUE))
					{
						// Store start point of string
						szStringSave         = szString;
						iState = C_KEY_VALUE;
					}
					// Handle CRLF after header
					else if(*szString == '\r')
					{
						// Store buffer
						sBuffer.append(szStringSave, szString - szStringSave);
						// Parse header
						if (iHeaderKeyId == C_KEY_CONTENT_DISPOSITION)
						{
							ParseHeader();
						}
						// Store content type
						else if (iHeaderKeyId == C_KEY_CONTENT_TYPE)
						{
							oFile.content_type = sBuffer;
						}
						else if (iHeaderKeyId == C_KEY_CONTENT_TRANSFER_ENCODING)
						{
							oFile.content_transfer_encoding = sBuffer;
						}
						// Clear buffer
						sBuffer.erase();

						// End of buffer?
						++szString;
HEADER_LF:
						if (szString == szStringEnd) { iState = C_HEADER_LF; return; }

						// LF after CR, REQUIRED in all cases!
						if(*szString != '\n') { iState = C_ERROR; return; }

						// End of buffer?
						++szString;
						if (szString == szStringEnd) { iState = C_HEADER_CR2; return; }

						// Store new buffer position
						szStringSave = szString;
						break;
					}
					// End of buffer?
					++szString;
					if (szString == szStringEnd) { return; }
				}
//fprintf(stderr, "'''''''''''''''''''''''''''''''''''''''''''''''''''''' `0x%02X` 0x%02X\n", *szString, '\r');
HEADER_CR2:
				// CRLF/CRLR matches end of section
				if(*szString == '\r')
				{
					++szString;
					if (szString == szStringEnd) { iState = C_HEADER_LF2; return; }
//fprintf(stderr, "C_HEADER_LF2\n");
HEADER_LF2:
					// LF after CR, REQUIRED in all cases!
					if(*szString != '\n') { iState = C_ERROR; return; }

					iState = C_BODY;
					// End of buffer?

					// Store new buffer position of body
					szStringSave = szString;
					++szStringSave;
//fprintf(stderr, "C_BODY\n");
					// Parse value
					break;
				}
				// Wait for next key
				else { iState = C_KEY_NAME; }
			}
		}
//fprintf(stderr, "===================> `0x%02X` 0x%02X\n", *szString, '\r');
		// End of string?
		++szString;
		if (szString == szStringEnd)
		{
			HandleEndOfBuffer(szStringSave, szStringEnd, iState);
			return;
		}
	}
}

//
// Get parser state
//
INT_32 MultipartParser::GetState() const
{
	switch (iState)
	{
		case C_ERROR:
			return -1;

		case C_END_MIME:
			return 1;

		default:
			return 0;
	}
}

//
// End of parsing process
//
void MultipartParser::ParseDone()
{
	;;
}

//
// Get error description
//
STLW::string MultipartParser::GetError() const { return ""; }

//
// Compare strings.
//
CCHAR_P MultipartParser::StrNFirstCaseStr(CCHAR_P sX, UINT_32 iMaxChars, CCHAR_P sY)
{
	while (tolower(*sX) == tolower(*sY))
	{
		if (iMaxChars == 0) { return NULL; }
		++sY; ++sX; --iMaxChars;

		if (*sY == '\0') { return sX; }
	}

return NULL;
}

//
// Compare strings.
//
CCHAR_P MultipartParser::StrFirstCaseStr(CCHAR_P sX, CCHAR_P sY)
{
	while (tolower(*sX) == tolower(*sY))
	{
		++sY; ++sX;
		if (*sY == '\0') { return sX; }
	}

return NULL;
}

//
// Compare strings
//
CCHAR_P MultipartParser::StrCaseStr(CCHAR_P sX, CCHAR_P sY)
{
	CCHAR_P szTMP = sX;
	while (*szTMP != '\0')
	{
		sX = szTMP;
		while (tolower(*sX) == tolower(*sY))
		{
			++sY; ++sX;
			if (*sY == '\0') { return sX; }
		}
		++szTMP;
	}

return NULL;
}

//
// A destructor
//
MultipartParser::~MultipartParser() throw()
{
	;;
}

} // namespace CAS
// End.
